// ########################################################################
// ## randcubic.cc:                                                      ##
// ## Klasse zur randomisierten Erzeugung kubischer Graphen              ##
// ## Copyright Thomas Dreibholz, uzs6pq@ibm.rhrz.uni-bonn.de            ##
// ##           WWW Homepage:     http://ibm.rhrz.uni-bonn.de/~uzs6pq    ##
// ########################################################################
 
#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>

#include "anzahltabelle.h"
#include "anzahltabelle.icc"

#include "../zufall/zufall.h"
#include "../tdgf/tdgf.h"
#include "../tdgf/graph.h"
#include "../tdgf/graph.icc"

#include "randcubic.h"


// #define DEBUG


#ifdef DEBUG
#define DB(text) { printf("DEBUG: %s  n=%d\n",text,n); for(int i = 1;i <= 3*n;i+= 2) { printf("(%d,%d) ",edges[i],edges[i + 1]); } puts(""); }
#define DB_T { if((int32)t < 0) { printf("t = %d - Abbruch!\n",t); exit(0); } }
#else
#define DB(text)
#define DB_T
#endif


// Globale Anzahltabelle
Anzahltabelle tabelle;


// Konstruktor
RandCubic::RandCubic(const card32 max)
{
   card32 i;

   // Klasse initialisieren
   knoten = max;
   // Wg. ungeraden Knotenwerten: 1 zustzliches Element am Ende
   labels    = new card32[knoten + 2];
   edges     = new card32[3 * knoten + 2];
   for(i = 0;i <= knoten + 1;i++)     labels[i] = 2000000+i;
   for(i = 0;i <= 3 * knoten + 1;i++) edges[i]  = 1000000+i;
   pos = 0;
}


// Destruktor
RandCubic::~RandCubic()
{
   delete labels;
   delete edges;
   labels = edges = NULL;
}


// Binre Ausgabe
void RandCubic::Schreibe(ostream& os, const bool bitMatrix,
                         const card32 typ, const card32 neg)
{
   card32 i;
   char   label[16];

   Graph graph(knoten,bitMatrix ? BIT_MATRIX : BYTE_MATRIX);
   for(i = 1;i <= 3 * knoten - 1;i +=  2) {
      if((edges[i] <= knoten) && (edges[i] > 0) &&
         (edges[i + 1] <= knoten) && (edges[i + 1] > 0)) {
         graph.Verbinde(edges[i] - 1, edges[i + 1] - 1);
      }
   }

   for(i = 1;i <= knoten;i++) {
      sprintf(label,"%d",i);
      if((typ == A_GRAPH) && (i == edges[3 * knoten - 1])) 
         sprintf(label,"%d*",i);
      else if(((typ == B_GRAPH) || (typ == C_GRAPH)) && (i == neg)) 
         sprintf(label,"%d-",i);
      else if(((typ == B_GRAPH) || (typ == C_GRAPH)) && (i == (edges[3 * knoten - 2])))
         sprintf(label,"%d+",i);
      graph.SetzeLabel(i - 1,label);
   }

   if(typ > 0) {
      switch(typ) {
         case 1:
            graph.SetzeTitel("A-Graph");
          break;
         case 2:
            graph.SetzeTitel("B-Graph");
          break;
         case 3:
            graph.SetzeTitel("C-Graph");
          break;
         case 4:
            graph.SetzeTitel("D-Graph");
          break;
      }
   }

   graph.Schreibe(os);
}


// Ausgabe-Operator
ostream& operator <<(ostream& os, const RandCubic& graph)
{
   card32 i,j,l1,l2,v,w;

   // Ausgabe der Labels
   for(i = 1;i <= graph.knoten;i++) {
      os << "K" << i << "=" << graph.labels[i] << "  ";
   }
   os << endl;

   // Ausgabe der Kanten durch Angabe der Labels
   os << "Kanten nach Label:        ";
   for(i = 1;i <= 3 * graph.knoten;i +=  2) {
      if((graph.edges[i] <= graph.knoten) && (graph.edges[i] > 0) &&
         (graph.edges[i + 1] <= graph.knoten) && (graph.edges[i + 1] > 0)) {
         os << "(" << graph.edges[i] << "," << graph.edges[i + 1]  << ") ";
      }
   }
   os << endl;

   // Ausgabe der Kanten durch Angabe der Knotennummern
   os << "Kanten nach Knotennummer: ";
   for(i = 1;i <= 3 * graph.knoten;i +=  2) {
      l1 = graph.edges[i];
      l2 = graph.edges[i + 1];
      v = w = 0;
      // Knotennummer fr Label l1 ermitteln
      for(j = 1;j <= graph.knoten;j++) {
         if(l1 == graph.labels[j]) {
            v = j;
            break;
         }
      }
      // Knotennummer fr Label l2 ermitteln
      for(j = 1;j <= graph.knoten;j++) {
         if(l2 == graph.labels[j]) {
            w = j;
            break;
         }
      }
      if((v != 0) && (w != 0))
         os << "(" << v << "," << w << ") ";
   }
   os << endl;

   return os;
}


// Choose-Funktion:
// Labels L[n-k+1] bis L[n] sollen eine zufllige, geordnete Auswahl
// der Labels L[1] bis L[n] sein.
void RandCubic::Choose(const card32 n, const card32 k)
{
   card32 x,i;
   int32  t;

   if(k >= n) return;
   for(i = n;i >= 1 + n - k;i--) {
      x = Random(i);
      t = labels[x];
      labels[x] = labels[i];
      labels[i] = t;
   }
}


// GraphA-Funktion:
// Erzeugung eines A-Graphen mit n Knoten basierend auf Lemma 1
// Voraussetzungen: n >=4 mu ungerade sein!
void RandCubic::GraphA(const card32 n)
{
   DB("RandCubic::GraphA() start");
   card32 neg = 87654321;
   int32  t;

   #ifdef DEBUG
      if(n < 4) {
         cerr << "RandCubic::GraphA(n) - Abbruch: n < 4" << endl;
         exit(0);
      }
      if((n % 2) == 0) {
         cerr << "RandCubic::GraphA(n) - ungerades n erwartet!" << endl;
         exit(0);
      }
   #endif

   Choose(n,1);
   GraphB(n - 1,neg);

   t = 3 * n - 5;
   DB_T
   pos = edges[t];
   // pos und neg sind die Labels des positiven und negativen Pols des
   // oben generieren B-Graphen.
   // Setzen der neuen Kanten:
   edges[t + 1] = neg;
   edges[t + 2] = labels[n];
   edges[t + 3] = pos;
   edges[t + 4] = labels[n];

   DB("RandCubic::GraphA() ende");
}


// GraphB-Funktion:
// Erzeugung eines B-Graphen mit n Knoten basierend auf Lemma 2 bis 8
// Voraussetzungen: n >=4 mu gerade sein!
void RandCubic::GraphB(const card32 n, card32& neg)
{
   DB("RandCubic::GraphB() start");
   AnzahlTyp x;

   #ifdef DEBUG
      if(n < 4) {
         cerr << "RandCubic::GraphB(n) - Abbruch: n < 4" << endl;
         exit(0);
      }
      if((n % 2) != 0) {
         cerr << "RandCubic::GraphB(n) - gerades n erwartet!" << endl;
         exit(0);
      }
   #endif

   x = Random(tabelle.B(n));
   if(x <= tabelle.C(n)) {
      GraphC(n,neg);
   }
   else {
      if(x <= tabelle.C(n) + (AnzahlTyp)(n * (n - 1) * (n - 2)) * tabelle.A(n - 3)) {
         GraphB2(n,neg);
      }
      else {
         GraphB3(n,neg);
      }
   }

   DB("RandCubic::GraphB() ende");
}


// GraphB2-Funktion:
// Erzeugung eines B-Graphen Typ 2 mit n Knoten basierend auf Lemma 2
void RandCubic::GraphB2(const card32 n, card32& neg)
{
   DB("RandCubic::GraphB2() start");
   card32 w,x;
   int32  t;

   Choose(n,3);
   GraphA(n - 3);
   w   = labels[n - 2];
   neg = labels[n - 1];
   pos = labels[n];
   t   = 3 * n - 10;
   DB_T
   x   = edges[t];
   // Die Kanten des oben generierten A-Graphen liegen in edges[1] bis
   // edges[t]. x ist der Label des Knotens mit Grad 2.
   // Setzen der neuen Kanten:
   edges[t + 1] = x;
   edges[t + 2] = w;
   edges[t + 3] = w;
   edges[t + 4] = neg;
   edges[t + 5] = w;
   edges[t + 6] = pos;
   edges[t + 7] = neg;
   edges[t + 8] = pos;

   DB("RandCubic::GraphB2() ende");
}


// GraphB3-Funktion:
// Erzeugung eines B-Graphen Typ 3 mit n Knoten basierend auf Lemma 3
void RandCubic::GraphB3(const card32 n, card32& neg)
{
   DB("RandCubic::GraphB3() start");
   card32 w,x;
   int32  t;

   Choose(n,2);
   GraphB(n - 2,neg);

   t   = 3 * n - 8;
   DB_T
   w   = edges[t];
   x   = neg;
   neg = labels[n - 1];
   pos = labels[n];
   // Im oben generierten Graph hat der positive Pol den Label w und
   // der negative Pol den Label x.
   // Setzen der neuen Kanten:
   edges[t + 1] = x;
   edges[t + 2] = neg;
   edges[t + 3] = w;
   edges[t + 4] = pos;
   edges[t + 5] = neg;
   edges[t + 6] = pos;

   DB("RandCubic::GraphB3() ende");
}


// GraphC-Funktion:
// Erzeugung eines C-Graphen mit n Knoten basierend auf Lemma 4 bis 8
// Voraussetzungen: n >=4 mu gerade sein!
void RandCubic::GraphC(const card32 n, card32& neg)
{
   DB("RandCubic::GraphC() start");
   AnzahlTyp r1,r2,r3,r4;
   AnzahlTyp x;

   #ifdef DEBUG
      if(n < 4) {
         cerr << "RandCubic::GraphC(n) - Abbruch: n < 4" << endl;
         exit(0);
      }
      if((n % 2) != 0) {
         cerr << "RandCubic::GraphC(n) - gerades n erwartet!" << endl;
         exit(0);
      }
   #endif

   x = Random(tabelle.C(n));

   r1 = ((3 * (AnzahlTyp)n - 8) / 2) * (AnzahlTyp)n * tabelle.A(n - 1);
   r2 = ((AnzahlTyp)n * ((AnzahlTyp)n - 1) * tabelle.B(n - 2)) + r1;
   if(n == 4)
      r3 = x;
   else
      r3 = r2 + (AnzahlTyp)(n * (n - 1) * (n - 2) * (n - 3)) * tabelle.D(n - 4) / 2;
   if(n <= 5)
      r4 = r3;
   else
      r4 = r3 + (AnzahlTyp)(n * (n - 1) * (n - 2) * (n - 3) * (n - 4)) * tabelle.A(n - 5) / 2;

   #ifdef DEBUG
      printf("RandGraph::GraphC() ->\n");
      printf("x    = %ld\n",x);
      printf("C[n] = %ld",tabelle.C(n)); printf("   (n = %d)\n",n);
      printf("r1   = %ld\n",r1);
      printf("r2   = %ld\n",r2);
      printf("r3   = %ld\n",r3);
      printf("r4   = %ld\n",r4);
      if(r4 > tabelle.C(n)) { puts("r4 > C[N]"); exit(1); }
      if(r3 > tabelle.C(n)) { puts("r3 > C[N]"); exit(1); }
      if(r2 > tabelle.C(n)) { puts("r2 > C[N]"); exit(1); }
      if(r1 > tabelle.C(n)) { puts("r1 > C[N]"); exit(1); }
   #endif

   // r[j] ist die Anzahl der C-Graphen mit n Knoten des Typs j.
   if(x <= r1)               GraphC1(n,neg);
   if((r1 < x) && (x <= r2)) GraphC2(n,neg);
   if((r2 < x) && (x <= r3)) GraphC3(n,neg);
   if((r3 < x) && (x <= r4)) GraphC4(n,neg);
   if(r4 < x)                GraphC5(n,neg);

   DB("RandCubic::GraphC() ende");
}


// GraphC1-Funktion:
// Erzeugung eines C-Graphen Typ 1 mit n Knoten basierend auf Lemma 4
void RandCubic::GraphC1(const card32 n, card32& neg)
{
   DB("RandCubic::GraphC1() start");
   card32 v,w,pos,x;
   int32  t;

   Choose(n,1);
   GraphA(n - 1);

   neg = edges[3 * n - 4];
   // neg ist der Label des Knotens mit Grad 2 im oben erzeugten A-Graphen

   x   = Random((3 * n - 8) / 2);
   v   = edges[2 * x - 1];
   w   = edges[2 * x];
   // v und w sind zwei zufllig gewhlte Kanten, die nicht mit dem Knoten
   // vom Grad 2 inzident sind.

   t = 3 * n - 6;
   DB_T
   edges[2 * x - 1] = edges[t + 1];
   edges[2 * x]     = edges[t + 2];
   // edges[1] bis edges[t] beinhalten jetzt die Kanten des A-Graphen ohne
   // die Kante v-w.
   // Setzen der neuen Kanten:
   pos = labels[n];
   edges[t + 1] = v;
   edges[t + 2] = pos;
   edges[t + 3] = w;
   edges[t + 4] = pos;

   DB("RandCubic::GraphC1() ende");
}


// GraphC2-Funktion:
// Erzeugung eines C-Graphen Typ 2 mit n Knoten basierend auf Lemma 5
void RandCubic::GraphC2(const card32 n, card32& neg)
{
   DB("RandCubic::GraphC2() start");
   card32 u,v,w,pos,x;
   int32  t;

   Choose(n,2);
   GraphB(n - 2,neg);
   //
   t   = 3 * n - 12;
   DB_T
   w   = edges[t + 1];
   x   = edges[t + 3];
   pos = edges[t + 4];
   u   = labels[n - 1];
   v   = labels[n];
   // pos ist der Label des positiven Pols des oben erzeugen B-Graphen.
   // w und x sind dessen Nachbarknoten. edges[1] bis edges[t] beinhalten
   // des B-Graphen.
   // Setzen der neuen Kanten:
   edges[t +  1] = w;
   edges[t +  2] = u;
   edges[t +  3] = x;
   edges[t +  4] = v;
   edges[t +  5] = u;
   edges[t +  6] = v;
   edges[t +  7] = u;
   edges[t +  8] = pos;
   edges[t +  9] = v;
   edges[t + 10] = pos;

   DB("RandCubic::GraphC2() ende");
}


// GraphC3-Funktion:
// Erzeugung eines C-Graphen Typ 3 mit n Knoten basierend auf Lemma 6
void RandCubic::GraphC3(const card32 n, card32& neg)
{
   DB("RandCubic::GraphC3() start");
   card32 u,v,pos;
   int32  t;

   Choose(n,4);
   if(n > 4)
      GraphD(n - 4);

   neg = labels[n - 3];
   u   = labels[n - 2];
   v   = labels[n - 1];
   pos = labels[n];
   t = 3 * n - 12;
   DB_T

   // Setzen der neuen Kanten:
   edges[t +  1] = neg;
   edges[t +  2] = u;
   edges[t +  3] = neg;
   edges[t +  4] = v;
   edges[t +  5] = u;
   edges[t +  6] = v;
   edges[t +  7] = u;
   edges[t +  8] = pos;
   edges[t +  9] = v;
   edges[t + 10] = pos;

   DB("RandCubic::GraphC3() ende");
}


// GraphC4-Funktion:
// Erzeugung eines C-Graphen Typ 4 mit n Knoten basierend auf Lemma 7
void RandCubic::GraphC4(const card32 n, card32& neg)
{
   DB("RandCubic::GraphC4() start");
   card32 u,v,w,pos;
   int32  t;

   Choose(n,5);
   GraphA(n - 5);

   neg = labels[n - 4];
   w   = labels[n - 3];
   u   = labels[n - 2];
   v   = labels[n - 1];
   pos = labels[n];
   t = 3 * n - 16;
   DB_T
   // Die Kanten des A-Graphen liegen in edges[1] bis edges[t]. edges[t] ist
   // der Label des Knotens vom Grad 2.
   // Setzen der neuen Kanten:
   edges[t +  1] = edges[t];
   edges[t +  2] = neg;
   edges[t +  3] = neg;
   edges[t +  4] = w;
   edges[t +  5] = w;
   edges[t +  6] = u;
   edges[t +  7] = w;
   edges[t +  8] = v;
   edges[t +  9] = u;
   edges[t + 10] = v;
   edges[t + 11] = u;
   edges[t + 12] = pos;
   edges[t + 13] = v;
   edges[t + 14] = pos;

   DB("RandCubic::GraphC4() ende");
}


// GraphC5-Funktion:
// Erzeugung eines C-Graphen Typ 5 mit n Knoten basierend auf Lemma 8
void RandCubic::GraphC5(const card32 n, card32& neg)
{
   DB("RandCubic::GraphC5() start");
   card32 u,v,w,pos,x;
   int32  t;

   Choose(n,4);
   GraphB(n - 4,neg);

   t = 3 * n - 14;
   DB_T
   x   = edges[t];
   w   = labels[n - 3];
   u   = labels[n - 2];
   v   = labels[n - 1];
   pos = labels[n];

   edges[t +  1] = x;
   edges[t +  2] = w;
   edges[t +  3] = w;
   edges[t +  4] = u;
   edges[t +  5] = w;
   edges[t +  6] = v;
   edges[t +  7] = u;
   edges[t +  8] = v;
   edges[t +  9] = u;
   edges[t + 10] = pos;
   edges[t + 11] = v;
   edges[t + 12] = pos;

   DB("RandCubic::GraphC5() ende");
}


// GraphD-Funktion:
// Erzeugung eines D-Graphen mit n Knoten basierend auf Lemma 9
// Voraussetzungen: n >=4 mu gerade sein!
void RandCubic::GraphD(const card32 n)
{
   DB("RandCubic::GraphD() start");
   card32 neg = 12345678;
   int32  t;

   GraphC(n,neg);
   t = 3 * n - 2;
   DB_T
   edges[t + 1] = neg;
   edges[t + 2] = edges[t];

   DB("RandCubic::GraphD() ende");
}


// Labels initialisieren
void RandCubic::InitialisiereLabels(const card32 n)
{
   for(card32 i = 1;i <= n;i++) labels[i] = i;
}


// RandCubic-Funktion:
// Generierung eines kubischen Graphen mit n Knoten starten; die Knotenzahl n
// wird beim Aufruf des Konstruktors angegeben!
void RandCubic::Start()
{
#ifdef DEBUG
   printf("RandCubic::Start() ->\n");
#endif

   InitialisiereLabels(knoten);
   GraphD(knoten);

#ifdef DEBUG
   printf("RandCubic::Start() beendet!\n\n\n");
#endif
}
