// ########################################################################
// ## graph.cc:                                                          ##
// ## Klasse zur Verwaltung von Graphen                                  ##
// ## Copyright Thomas Dreibholz, uzs6pq@ibm.rhrz.uni-bonn.de            ##
// ##           WWW Homepage:     http://ibm.rhrz.uni-bonn.de/~uzs6pq    ##
// ########################################################################

#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <string.h>

#include "graph.h"
#include "graph.icc"

#include "../tdgf/tdgf.h"


#define DEBUG


// ########################################################################
// ## Konstruktoren und Destruktor                                       ##
// ########################################################################


// Konstruktor 0
Graph::Graph()
{
   adjazenzMatrix = NULL;
   nachfolger     = NULL;
   knoten         = 0;
   flags          = 0;
   titel          = NULL;
   labels         = NULL;
   biparts        = NULL;
}


// Klasse initialisieren
void Graph::Initialisiere(const card32 n, const card16 flgs)
{
   card32 i;

   knoten     = (card16)n;
   kanten     = 0;
   flags      = flgs;
   titel      = NULL;
   labels     = NULL;
   biparts    = NULL;
   nachfolger = NULL;
   adjazenzMatrix = new card8[AdjazenzBytes(n,flags)];
   if(adjazenzMatrix == NULL) {
      cerr << "Nicht genug freier Speicher!" << endl;
      exit(1);
   }
   for(i = 0;i < AdjazenzBytes(n,flags);i++) {
      adjazenzMatrix[i] = 0;
   }
}


// Klasse freigeben
void Graph::GibFrei()
{
   if(adjazenzMatrix != NULL) {
      delete adjazenzMatrix;
      adjazenzMatrix = NULL;
      nachfolger     = NULL;
      if(titel != NULL) {
         delete titel;
         titel = NULL;
      }
      if(labels != NULL) {
         delete labels;
         labels = NULL;
      }
      if(biparts != NULL) {
         delete biparts;
         biparts = NULL;
      }
   } 
}


// Konstruktor 1
Graph::Graph(const card32 n, const card16 flgs)
{  
   Initialisiere(n,flgs);
}


// Konstruktor 2
Graph::Graph(const Graph& original)
{  
   Clone(original);
}


// Copy-Konstruktor
void Graph::operator=(const Graph& original)
{  
   GibFrei();
   Clone(original);
}


// Clone-Operation
void Graph::Clone(const Graph& original)
{
   card32 i;

   Initialisiere(original.knoten,original.flags);
   kanten = original.kanten;
   flags  = original.flags;
   if(original.titel != NULL) {
      titel = new char[strlen(original.titel) + 1];
      strcpy(titel,original.titel);
   }
   if(original.labels != NULL) {
      labels = new char[4 * knoten];
      memcpy(labels,original.labels,4 * knoten);
   }
   if(original.biparts != NULL) {
      biparts = new card8[(knoten + 7) / 8];
      memcpy(biparts,original.biparts,(knoten + 7) / 8);
   }
   for(i = 0;i < AdjazenzBytes(original.knoten,original.flags);i++) {
      adjazenzMatrix[i] = original.adjazenzMatrix[i];
   }  
}


// Destruktor
Graph::~Graph()
{
   GibFrei();
}


// Graph neu initialisieren
void Graph::Neu()
{
   card32 n = knoten;
   GibFrei();
   Initialisiere(n,flags);
}


// ########################################################################
// ## ndern von Einstellungen                                           ##
// ########################################################################


// Titel des Graphen setzen
void Graph::SetzeTitel(const char* neuerTitel)
{
   if(titel != NULL) {
      delete titel;
      titel = NULL;
   }
   if(neuerTitel != NULL) {
      titel = new char[strlen(neuerTitel) + 1];
      strcpy(titel,neuerTitel);
   }
}


// Labels des Graphen setzen
void Graph::SetzeLabels(const char* neueLabels)
{
   if(labels != NULL) {
      delete labels;
      labels = NULL;
   }
   if(neueLabels != NULL) {
      labels = new char[4 * knoten];
      memcpy(labels,neueLabels,4 * knoten);
   }
}


// Label des Graphen abfragen
bool Graph::Label(const card32 n, char* speicher) const
{
   card32 i;

#ifdef DEBUG
   if(n >= knoten) {
      cerr << "Graph::Label() - Unzulssige Knotennummer!" << endl;
      exit(1);
   }
#endif

   for(i = 0;i < 5;i++) speicher[i] = 0x00;
   if(labels != NULL) {
      for(i = 0;i < 4;i++)
         speicher[i] = labels[(4 * n) + i];
      return(speicher[0] != 0x00);
   }
   return(false);
}


// Label setzen
void Graph::SetzeLabel(const card32 n, const char* label)
{
   card32 i;

#ifdef DEBUG
   if(n >= knoten) {
      cerr << "Graph::SetzeLabel() - Unzulssige Knotennummer!" << endl;
      exit(1);
   }
#endif

   if(labels == NULL) {
      labels = new char[4 * knoten];
      for(i = 0;i < 4 * (card32)knoten;i++) labels[i] = 0x00;
   }
   for(i = 0;i < 4;i++)
      labels[(4 * n) + i] = label[i];
}


// Bipart-Einstellungen des Graphen setzen
void Graph::SetzeBiparts(const card8* neueBiparts)
{
   if(biparts != NULL) {
      delete biparts;
      biparts = NULL;
   }
   if(neueBiparts != NULL) {
      biparts = new card8[(knoten + 7) / 8];
      memcpy(biparts,neueBiparts,(knoten + 7) / 8);
   }
}


// Bipart-Eintrag des Graphen abfragen
bool Graph::Bipart(const card32 n) const
{
#ifdef DEBUG
   if(n >= knoten) {
      cerr << "Graph::Label() - Unzulssige Knotennummer!" << endl;
      exit(1);
   }
#endif

   if(biparts != NULL) {
      return(biparts[n / 8] & (1 << (n % 8)));
   }
   return(false);
}


// Bipart-Eintrag des Graphen setzen
void Graph::SetzeBipart(const card32 n, const bool menge)
{
   card32 i;

#ifdef DEBUG
   if(n >= knoten) {
      cerr << "Graph::SetzeBipart() - Unzulssige Knotennummer!" << endl;
      exit(1);
   }
#endif

   if(biparts == NULL) {
      biparts = new card8[(knoten + 7) / 8];
      for(i = 0;i < ((card32)knoten + 7) / 8;i++) biparts[i] = 0x00;
   }
   i = (n / 8);
   if(menge)
      biparts[i] |= (1 << (n % 8));
   else
      biparts[i] &= ~(1 << (n % 8));
}


// ########################################################################
// ## Verwaltung der Adjazenzmatrix                                      ##
// ########################################################################


// Kante zwischen Knoten a und b hinzufgen oder lschen
void Graph::SetzeKante(const card32 a, const card32 b,
                          const bool neueKante)
{
#ifdef DEBUG
   if((a >= knoten) || (b >= knoten)) {
      cerr << "Graph::SetzeKante() - Unzulssige Knotennummern!" << endl;
      exit(1);
   }
   if(a == b) {
      cerr << "Graph::SetzeKante() - Schleifen sind nicht erlaubt!" << endl;
      exit(1);
   }
   if(Verbunden(a,b) == neueKante) {
      cerr << "Graph::Kante() - Befehl unntig!" << endl;
      exit(1);
    }
#endif

   if(flags & BIT_MATRIX) {
      // Bitposition in der Adjazenzmatrix berechnen
      const card32 z         = min(a,b);
      const card32 s         = max(a,b);
      const card32 bitNummer = (z * knoten) + s - ((z + 2) * (z + 1)) / 2;
      const card32 byte      = bitNummer / 8;
      const card32 bit       = bitNummer % 8;

      // Neue Kante hinzufgen bzw. entfernen
      if(neueKante) {
         if(! (adjazenzMatrix[byte] & (1 << bit))) {
            adjazenzMatrix[byte] += (1 << bit);
            kanten++;
         }
      }
      else {
         if(adjazenzMatrix[byte] & (1 << bit)) {
            adjazenzMatrix[byte] -= (1 << bit);
            kanten--;
         }
      }
   }
   else {
      const card32 z = min(a,b);
      const card32 s = max(a,b);
      adjazenzMatrix[(z * knoten) + s - (((z + 2) * (z + 1)) / 2)] = (char)neueKante;
      if(neueKante)
         kanten++;
      else
         kanten--;
   }
}


// ########################################################################
// ## Ausgabe                                                            ##
// ########################################################################


// Output-Operator
ostream& operator<<(ostream& os, const Graph& graph)
{
   card32 i,j;

   if(graph.knoten > 0) {
      os << "Knoten   = " << graph.knoten << endl; 
      os << "Kanten   = " << graph.kanten << endl;
      if(graph.titel != NULL)
         os << "Titel    = " << graph.titel << endl;
      if(graph.labels != NULL) {
         char label[16];

         os << "Labels   = ";
         for(i = 0;i < graph.knoten;i++) {
            os << "K" << i + 1 << "=";
            if(graph.Label(i,label)) {
               os << label << " ";
            }
         }
         os << endl;
      }
      if(graph.biparts != NULL) {
         os << "Bipart   = rot:  ";
         for(i = 0;i < graph.knoten;i++) {
             if(graph.Bipart(i)) cout << i << "  ";
         }
         os << endl;
         os << "           blau: ";
         for(i = 0;i < graph.knoten;i++) {
             if(!(graph.Bipart(i))) cout << i << "  ";
         }
         os << endl;
      }
      os << "AdMatrix =   ";
      for(i = 0;i < graph.knoten;i++) {
         os.width(3);
         os << i;
      }
      os << endl << "          ";
      for(card32 zeile = 0;zeile < graph.knoten;zeile++) {
         os.width(3);
         os << zeile << " ";
         for(card32 spalte = 0;spalte < graph.knoten;spalte++) {
            if(zeile == spalte) 
               os << " - ";
            else if(graph.Verbunden(zeile,spalte))
               os << " 1 ";
            else
               os << " 0 ";
         }
         os << endl;
         if(zeile < (card32)graph.knoten - 1) {
            os << "          ";
         }       
      }
      os << "Verbindg = ";    
      for(i = 0; i < graph.knoten;i++) {
         for(j = i + 1; j < graph.knoten;j++) {
            if(graph.Verbunden(i,j)) {
               os << i << "-" << j << "  ";       
            }
         }
      }
      os << endl;                            
   }
   else { 
      os << "( leerer Graph )" << endl;
   }

   return(os);
}



// ########################################################################
// ## Lesen und Schreiben                                                ##
// ########################################################################


// Graph aus TDGF-Datei lesen
bool Graph::Lese(istream& is)
{
   char        id[4];
   card16      i;
   GRDT_Header grdt;

   // Andere Datenstze berspringen
   is.read((void *)&id,4);
   while(is.good()) {     
      is.seekg(-4,ios::cur);

      // Neuer Graph -> weiter
      if(!(strncmp(id,GRDT_Header_ID,4))) {
         break;        
      }

      // Sonstige Header -> berspringen
      else {
         is.seekg(4,ios::cur);
         is.read((void *)&i,2);
         is.seekg(i,ios::cur);         
      }

      // Nchste ID lesen
      is.read((void *)&id,4);      
   }

   // GRDT-Header lesen und verarbeiten
   is.read((void *)&grdt,sizeof(grdt));
   if((is.fail()) || (strncmp(grdt.ID,GRDT_Header_ID,4))) return(false);
   knoten = grdt.Knoten;
   kanten = grdt.Kanten;
   flags  = grdt.Flags;

   // Adjazenzmatrix lesen
   if(adjazenzMatrix != NULL) 
      delete adjazenzMatrix;
   if(titel != NULL) {
      delete titel;
      titel = NULL;
   }
   if(labels != NULL) {
      delete labels;
      labels = NULL;
   }
   if(biparts != NULL) {
      delete biparts;
      biparts = NULL;
   }
   adjazenzMatrix = new card8[AdjazenzBytes(knoten,flags)];
   if(adjazenzMatrix == NULL) {
      cerr << "Graph::Lese() - Nicht genug freier Speicher!" << endl;
      exit(1);
   }
   is.read((void *)adjazenzMatrix,AdjazenzBytes(knoten,flags));  


   // Daten einer eventuellen spteren Version einfach berspringen
   is.seekg(grdt.Bytes -
           (sizeof(GRDT_Header) - 8 + AdjazenzBytes(knoten,flags)),ios::cur);
   if(is.fail()) return(false);


   // Weitere Datenstze verarbeiten
   is.read((void *)&id,4);
   while(is.good()) {     
      is.seekg(-4,ios::cur);

      // Neuer Graph -> zurck
      if(!(strncmp(id,GRDT_Header_ID,4))) {
         return(true);        
      }

      // GRTT-Header -> Titel lesen
      else if(!(strncmp(id,GRTT_Header_ID,4))) {
         GRTT_Header grtt;

         is.read((void *)&grtt,sizeof(grtt));
         titel = new char[grtt.Bytes];
         is.read((void *)titel,grtt.Bytes);

         // Weitere Daten einer eventuellen spteren Version einfach berspringen
         is.seekg(grtt.Bytes - sizeof(GRTT_Header) + 8 - strlen(titel) - 1,ios::cur);
      }

      // GRLD-Header -> Labels lesen
      else if(!(strncmp(id,GRLD_Header_ID,4))) {
         GRLD_Header grld;
         
         is.read((void *)&grld,sizeof(grld));
         labels = new char[grld.Bytes];
         is.read((void *)labels,grld.Bytes);

         // Weitere Daten einer eventuellen spteren Version einfach berspringen
         is.seekg(grld.Bytes - sizeof(GRLD_Header) + 8 - knoten * 4,ios::cur);
      }    

      // GRBD-Header -> Biparts lesen
      else if(!(strncmp(id,GRBD_Header_ID,4))) {
         GRBD_Header grbd;
         
         is.read((void *)&grbd,sizeof(grbd));
         biparts = new card8[grbd.Bytes];
         is.read((void *)biparts,grbd.Bytes);

         // Weitere Daten einer eventuellen spteren Version einfach berspringen
         is.seekg(grbd.Bytes - sizeof(GRBD_Header) + 8 - ((knoten + 7) / 8),ios::cur);
      }    

      // Sonstige Header -> berspringen
      else {
         is.seekg(4,ios::cur);
         is.read((void *)&i,2);
         is.seekg(i,ios::cur);         
      }

      // Nchste ID lesen
      is.read((void *)&id,4);      
   }

   return(true);   
}


// Graph in TDGF-Datei schreiben
bool Graph::Schreibe(ostream& os)
{
   SchreibeGRDT(os,flags,knoten,kanten,AdjazenzBytes(knoten,flags));
   os.write((void *)adjazenzMatrix,AdjazenzBytes(knoten,flags));
   if(titel != NULL) SchreibeGRTT(os,titel);
   if(labels != NULL) SchreibeGRLD(os,knoten,labels);
   if(biparts != NULL) SchreibeGRBD(os,knoten,biparts);
   return(os.fail());
}
