// ########################################################################
// ## tdgf.cc                                                            ##
// ## Prozeduren zum Bearbeiten von Dateien im TDGF-Format               ##
// ## Copyright Thomas Dreibholz, uzs6pq@ibm.rhrz.uni-bonn.de            ##
// ##           WWW Homepage:     http://ibm.rhrz.uni-bonn.de/~uzs6pq    ##
// ########################################################################

#include <string.h>
#include <fstream.h>
#include <time.h>

#include "../system.h"
#include "tdgf.h"


// TDGF-Datei erstellen
bool ErstelleTDGF(ofstream& os, const char* name,
                  const char* titel,
                  const char* copyright, const char* programm)
{
   os.open(name,ios::binary | ios::out);
   if(os.good()) {
      SchreibeTDGF(os);
      SchreibeINFO(os,titel,copyright,programm);
      return(true);
   }
   return(false);
}


// TDGF-Datei zum Lesen ffnen
tdgfFehler OeffneTDGF(ifstream& is, const char* name,
                      TDGF_Header* tdgf, INFO_Header* info)
{
   is.open(name,ios::binary | ios::in);
   if(is.good()) {
      // TDGF-Header lesen und prfen
      is.read((void *)tdgf,sizeof(TDGF_Header));
      if(strncmp(tdgf->ID,TDGF_Header_ID,4)) {
         is.close();
         return(format);
      }
      if(tdgf->ByteOrderTest != TDGF_Header_ByteOrderTest) {
         is.close();
         return(byteOrder);
      }
      if(tdgf->Kompression != TDGF_Header_KeineKompression) {
         is.close();
         return(kompression);
      }

      // Weitere Daten einer eventuellen spteren Version einfach berspringen
      is.seekg(tdgf->Bytes - (sizeof(TDGF_Header) - 8),ios::cur);
      
      // INFO-Header laden und prfen
      is.read((void *)info,sizeof(INFO_Header));
      if(strncmp(info->ID,INFO_Header_ID,4)) {
         is.close();
         return(format);
      } 
 
      // Weitere Daten einer eventuellen spteren Version einfach berspringen
      is.seekg(info->Bytes - (sizeof(INFO_Header) - 8),ios::cur);

      if(is.good()) return(OK);
   }
   return(dos);
}


// TDGF-Datensatz und INFO-Datensatz schreiben
bool SchreibeTDGF(ostream& os)
{
   struct TDGF_Header tdgf;

   // TDGF-Datensatz schreiben
   memcpy(&tdgf.ID,TDGF_Header_ID,4);
   tdgf.Bytes         = sizeof(tdgf) - 8;
   tdgf.ByteOrderTest = TDGF_Header_ByteOrderTest;
   tdgf.Kompression   = TDGF_Header_KeineKompression;
   tdgf.Version       = TDGF_Header_Version;

   os.write((void *)&tdgf,sizeof(TDGF_Header));
   return(os.fail());
}


// INFO-Datensatz und INFO-Datensatz schreiben
bool SchreibeINFO(ostream& os, const char* titel, const char* copyright,
                               const char* programm)
{
   struct INFO_Header info;
   time_t             systemZeit;
   tm*                zeit;
   int32              i;

   // Default-Werte
   if(titel == NULL)     titel = "";
   if(copyright == NULL) copyright = "";
   if(programm == NULL)  programm = "?";

   // Info- schreiben
   memcpy(&info.ID,INFO_Header_ID,4);
   info.Bytes = sizeof(info) - 8;
   for(i = 0;i < INFO_Header_StringLaenge;i++) {  
      info.Titel[i] = info.Copyright[i] = info.Programm[i] = 0x00;
   }
   strncpy(info.Titel,    titel,    min((int)strlen(titel),    INFO_Header_StringLaenge - 1));
   strncpy(info.Copyright,copyright,min((int)strlen(copyright),INFO_Header_StringLaenge - 1));
   strncpy(info.Programm, programm ,min((int)strlen(programm), INFO_Header_StringLaenge - 1));
   time(&systemZeit);
   zeit = localtime(&systemZeit);
   info.Tag     = zeit->tm_mday;
   info.Monat   = zeit->tm_mon + 1;
   info.Jahr    = zeit->tm_year + 1900;
   info.Stunde  = zeit->tm_hour;
   info.Minute  = zeit->tm_min;
   info.Sekunde = zeit->tm_sec;
   strncpy(info.Zone,tzname[0],3);
   info.Zone[3] = 0x00;   
   info.null[0] = 0;
   info.null[1] = 0;
   info.null[2] = 0;
   info.null[3] = 0;
   info.null[4] = 0;

   os.write((void *)&info,  sizeof(INFO_Header));     
   return(os.fail());
}


// GRDT-Datensatz und weitere Informationen schreiben
bool SchreibeGRDT(ostream& os, card16 flags,
                               card16 knoten,
                               card16 kanten,
                               card16 adjazenzBytes)
{
   GRDT_Header grdt;
   memcpy(&grdt.ID,GRDT_Header_ID,4);
   grdt.Bytes  = sizeof(grdt) - 8 + adjazenzBytes;
   grdt.Flags  = flags;
   grdt.Knoten = knoten;
   grdt.Kanten = kanten;
   grdt.null[0] = 0;
   grdt.null[1] = 0;
   os.write((void *)&grdt,sizeof(GRDT_Header));
   return(os.fail());
}


// GRTT-Datensatz schreiben
bool SchreibeGRTT(ostream& os, const char* titel)
{
   GRTT_Header grtt;
   memcpy(&grtt.ID,GRTT_Header_ID,4);
   grtt.Bytes = sizeof(grtt) - 8 + strlen(titel) + 1;
   os.write((void *)&grtt,sizeof(GRTT_Header));
   os.write((void *)titel,strlen(titel) + 1);
   return(os.fail());
}


// GRLD-Datensatz schreiben
bool SchreibeGRLD(ostream& os, card16 knoten, char* labels)
{
   GRLD_Header grld;
   memcpy(&grld.ID,GRLD_Header_ID,4);
   grld.Bytes = sizeof(grld) - 8 + knoten * 4;
   os.write((void *)&grld,sizeof(GRLD_Header));
   os.write((void *)labels,knoten * 4);
   return(os.fail());
}


// GRBD-Datensatz schreiben
bool SchreibeGRBD(ostream& os, card16 knoten, card8* biparts)
{
   GRBD_Header grbd;
   memcpy(&grbd.ID,GRBD_Header_ID,4);
   grbd.Bytes = sizeof(grbd) - 8 + ((knoten + 7) / 8);
   os.write((void *)&grbd,sizeof(GRBD_Header));
   os.write((void *)biparts,(knoten + 7) / 8);
   return(os.fail());
}
